#ifndef _EARLEY_PARSER_H_
#define _EARLEY_PARSER_H_

#include <vector>
#include <set>

#include <boost/iterator/indirect_iterator.hpp>

#include <outilex/compare_by_idx.h>
#include <outilex/ugrammar.h>
#include <outilex/uchart.h>
#include <outilex/uparser.h>


#if 0
template<typename RandomAccessContainer>
struct compare_by_idx {

  typedef RandomAccessContainer container_type;

  compare_by_idx(const container_type & C_) : pC(& C_) {
    //std::cerr << "compare_by_idx(& C = " << pC << ")\n";
  }

  bool operator()(int a, int b) const {
    return (*pC)[a] < (*pC)[b];
  }

  const container_type * pC; 
};

#endif

class earley_parser : public uparser {

public:

  typedef ugrammar::input_type  input_type;
  typedef ugrammar::output_type output_type;


  earley_parser(const ugrammar & grm) : gram(grm), pchart(NULL), agenda() {}
  ~earley_parser() {}


  void parse(uchart & chart, bool surf = false, double timeout = 0.);


public:

  /* internal data structures */

  struct e_item {

    const usyntagm_pattern * pfst;
    int q;
    int from, to;
    mutable double w;
    featstruct fs;
    synt_path_type path;

    e_item(const usyntagm_pattern & fst_, int q_, int from_, int to_,
           double w_, const featstruct & fs_, const synt_path_type & path_)
      : pfst(& fst_), q(q_), from(from_), to(to_), w(w_), fs(fs_), path(path_) {}


    // compare by weight ??
    bool operator<(const e_item & item) const {
      if (pfst != item.pfst) { return pfst < item.pfst; }
      if (q != item.q) { return q < item.q; }
      if (from != item.from) { return from < item.from; }
      if (to != item.to) { return to < item.to; }
      return path < item.path;
    }

    void dump(std::ostream & os) const;
  };


  struct earley_stack {

    earley_stack() : items(), item_set() {}

    earley_stack(const earley_stack & s) 
      : items(s.items), item_set(s.item_set.begin(), s.item_set.end()) {
        //std::cerr << "earley_stack:: copy constructor\n";
      }

    typedef std::multiset<e_item> item_set_type;
    typedef std::vector<item_set_type::iterator> vector_type;

    typedef boost::indirect_iterator<vector_type::iterator> iterator;
    typedef boost::indirect_iterator<vector_type::const_iterator> const_iterator;

    typedef item_set_type::iterator by_val_iterator;
    typedef item_set_type::const_iterator const_by_val_iterator;


    int size() const { return items.size(); }

    void add(const e_item & item);


    const e_item & operator[](int idx) const { return *(items[idx]); }

    iterator begin() { return iterator(items.begin()); }
    const_iterator begin() const { return const_iterator(items.begin()); }

    iterator end() { return iterator(items.end()); }
    const_iterator end() const { return const_iterator(items.end()); }

    by_val_iterator by_val_begin() { return item_set.begin(); }
    by_val_iterator by_val_end() { return item_set.end(); }

#if 0
    const_by_val_iterator lower_bound(const e_item & it) const {
      return item_set.lower_bound(it);
    }
#endif

    vector_type items;
    item_set_type item_set;
  };

  typedef std::vector<earley_stack> e_agenda;

  const ugrammar & gram;
  uchart * pchart;
  e_agenda agenda;

protected:

  void ENQUEUE(int pos, const e_item & item);
  void PREDICTOR(int pos, const std::string & syntname);
  void COMPLETER(int pos, const e_item & item);
};

inline std::ostream & operator<<(std::ostream & os, const earley_parser::e_item & item) {
  item.dump(os); return os;
}
#endif
